/*
 *   Copyright (c) 2004, Steve Dekorte
 *   All rights reserved. See _BSDLicense.txt.
 */

#include "DBFile.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* for open flags */
#include <fcntl.h>

/* for chmod mode */
#include <sys/types.h>
#include <sys/stat.h>

DBFile *DBFile_new(void)
{
  DBFile *self = calloc(1, sizeof(DBFile));
  return self;
}

void DBFile_free(DBFile *self)
{
  DBFile_close(self);
  if (self->path) free(self->path);
  if (self->lockPath) free(self->lockPath);
  free(self);
}

void DBFile_path_(DBFile *self, char *path)
{
  self->path = strcpy(realloc(self->path, strlen(path)+1), path);
  self->lockPath = strcpy(realloc(self->lockPath, strlen(path)+2), path);
  self->lockPath[strlen(path)] = '~';
  self->lockPath[strlen(path)+1] = 0x0;
}

char *DBFile_path(DBFile *self) { return self->path; }

void DBFile_open(DBFile *self)
{
  if (!self->db) 
  {
    self->db = dbopen(self->path, 
      O_CREAT | O_RDWR, 
      S_IRWXU | S_IRWXG | S_IRWXO, 
      DB_BTREE, NULL);
  }
}

void DBFile_close(DBFile *self)
{
  if (self->db) self->db->close(self->db);
  self->db = 0x0;
}

void DBFile_delete(DBFile *self)
{
  remove(self->path);
}

void DBFile_beginTransaction(DBFile *self)
{
  DBFile_lock(self);
  DBFile_open(self);
}

void DBFile_endTransaction(DBFile *self)
{
  self->db->sync(self->db, NULL);
  DBFile_unlock(self);
}

void DBFile_put(DBFile *self, DBT *key, DBT *data)
{
  DBFile_beginTransaction(self);
  self->db->put(self->db, key, data, NULL);
  DBFile_endTransaction(self);
}

void DBFile_get(DBFile *self, DBT *key, DBT *data)
{
  self->db->get(self->db, key, data, NULL);
}

void DBFile_del(DBFile *self, DBT *key)
{
  DBFile_beginTransaction(self);
  self->db->del(self->db, key, NULL);
  DBFile_endTransaction(self);
}

void DBFile_performEntry_(DBFile *self, DBEntry *entry)
{
  switch (entry->op)
  {
    case DBEntry_PUT:
      self->db->put(self->db, (void *)entry->key, (void *)entry->value, NULL);
      break;
    case DBEntry_DELETE:
      self->db->del(self->db, (void *)entry->key, NULL);
      break;
  }
}

/* -- lock ------------------------ */

void DBFile_lock(DBFile *self)
{
  FILE *fp = fopen(self->lockPath, "w");
  fclose(fp);
}

void DBFile_unlock(DBFile *self)
{
  remove(self->lockPath);
}

unsigned char DBFile_hasLock(DBFile *self)
{
  FILE *fp = fopen(self->lockPath, "r");
  if (fp) fclose(fp);
  return fp ? 1 : 0;
}

void DBFile_copyFromPath_(DBFile *self, char *path)
{
  DBFile_delete(self);
  {
  size_t size = 1024*32;
  unsigned char *buf = malloc(size);
  FILE *in  = fopen(path, "r");
  FILE *out = fopen(self->path, "w");
  for (;;)
  {
    size_t readSize = fread(buf, 1, size, in);
    if (readSize == 0) break;
    fwrite(buf, 1, readSize, out);
  }
  fclose(out);
  fclose(in);
  free(buf);
  }
  DBFile_unlock(self);
}


