/*
 *   Copyright (c) 2004, Steve Dekorte
 *   All rights reserved. See _BSDLicense.txt.
 */
 
#include "TwinDB.h"
#include <stdlib.h>
#include <string.h>

TwinDB *TwinDB_new(void)
{
  TwinDB *self = calloc(1, sizeof(TwinDB));
  self->db1 = DBFile_new();
  self->db2 = DBFile_new();
  self->writes = List_new();
  return self;
}

void TwinDB_free(TwinDB *self)
{
  TwinDB_close(self); /* will push and free remaining writes */
  DBFile_free(self->db1);
  DBFile_free(self->db2);
  List_free(self->writes);
  free(self);
}

void TwinDB_path_(TwinDB *self, char *path)
{
  self->path = strcpy(realloc(self->path, strlen(path)+1), path);
}

void TwinDB_setPaths(TwinDB *self)
{
  char *p1 = strcpy(malloc(strlen(self->path)+2), self->path);
  char *p2 = strcpy(malloc(strlen(self->path)+2), self->path);
  strcat(p1, "1.tdb");
  strcat(p2, "2.tdb");
  DBFile_path_(self->db1, p1);
  DBFile_path_(self->db2, p2);
  free(p1);
  free(p2);
}

void TwinDB_setWriteCacheSizeLimit_(TwinDB *self, size_t size)
{ self->writeCacheSizeLimit = size; }

void TwinDB_fixCorruptFiles(TwinDB *self)
{  
  if (DBFile_hasLock(self->db1))
  { 
    printf("%s corrupted, replacing with %s\n", 
      DBFile_path(self->db1), DBFile_path(self->db2)); 
    DBFile_copyFromPath_(self->db1, DBFile_path(self->db2)); 
  }
  if (DBFile_hasLock(self->db2))
  { 
    printf("%s corrupted, replacing with %s\n", 
      DBFile_path(self->db2), DBFile_path(self->db1));   
    DBFile_copyFromPath_(self->db2, DBFile_path(self->db1)); 
  }
}
  
void TwinDB_open(TwinDB *self)
{
  TwinDB_setPaths(self);
  TwinDB_fixCorruptFiles(self);
}

void TwinDB_close(TwinDB *self)
{
  TwinDB_sync(self);
  DBFile_close(self->db1); 
  DBFile_close(self->db2); 
}


void TwinDB_delete(TwinDB *self)
{ 
  TwinDB_close(self);
  DBFile_delete(self->db1); 
  DBFile_delete(self->db2); 
}


void TwinDB_sync(TwinDB *self)
{
  /* perform any cached writes */
  if (List_count(self->writes) == 0) return;
  /* reverse so we perform the oldest first */
  List_reverse(self->writes);
  
  DBFile_beginTransaction(self->db1);
  List_target_do_(self->writes, self->db1, (ListDoWithCallback *)DBFile_performEntry_);
  DBFile_endTransaction(self->db1);
  
  DBFile_beginTransaction(self->db2);
  List_target_do_(self->writes, self->db2, (ListDoWithCallback *)DBFile_performEntry_);
  DBFile_endTransaction(self->db2); 
  
  /* free the write entries and empty the cache */
  List_do_(self->writes, (ListDoCallback *)DBEntry_free);
  List_empty(self->writes);
  self->writeCacheSize = 0;
}

void TwinDB_addEntry_(TwinDB *self, DBEntry *entry)
{
  List_add_(self->writes, entry);
  self->writeCacheSize += DBEntry_size(entry);
  if (self->writeCacheSizeLimit && self->writeCacheSize > self->writeCacheSizeLimit)
  { TwinDB_sync(self); }
}

void TwinDB_put(TwinDB *self, DBT *key, DBT *data)
{  
  DBEntry *entry = DBEntry_new();
  DBEntry_set(entry, DBEntry_PUT, key, data);
  TwinDB_addEntry_(self, entry);
}

void TwinDB_get(TwinDB *self, DBT *key, DBT *data)
{
  TwinDB_sync(self);
  DBFile_get(self->db1, key, data);
}

void TwinDB_del(TwinDB *self, DBT *key)
{
  DBEntry *entry = DBEntry_new();
  DBEntry_set(entry, DBEntry_DELETE, key, 0x0);
  TwinDB_addEntry_(self, entry);
}



  
